import os, pydoc, inspect, urllib, types
import XmlFormatter

class ApiFormatter(XmlFormatter.XmlFormatter):

    def __init__(self, command, modules):
        XmlFormatter.XmlFormatter.__init__(self, command)
        self.documented_modules = modules
        return

    def ispublic(self, name):
        if hasattr(self.module, '__all__'):
            return name in self.module.__all__ and 'yes' or 'no'
        else:
            return not name.startswith('_') and 'yes' or 'no'

    def isdocumented(self, name):
        return self.documented_modules.has_key(name) and 'yes' or 'no'

    def doc_module(self, module):
        """
        Produce documentation for a given module object.
        """
        if not inspect.ismodule(module):
            raise TypeError("doc_module() requires a 'module' object "
                            "but received a '%s'" % type(module).__name__)

        name = module.__name__
        try:
            file = inspect.getabsfile(module)
            file = urllib.pathname2url(file)
        except TypeError:
            file = '<builtin>'

        if hasattr(module, '__path__'):
            root_element = 'package'
        else:
            root_element = 'module'

        attributes = {'name' : name,
                      'file' : file,
                      }
        self.start_element(root_element, attributes)

        desc = self.escape(pydoc.getdoc(module))
        self.write_element('description', content=desc)

        # standard module metadata
        for attr in ['author', 'credits', 'date', 'version']:
            if hasattr(module, '__%s__' % attr):
                content = self.escape(str(getattr(module, '__%s__' % attr)))
                self.write_element(attr, content=content)

        contents = []
        if hasattr(module, '__path__'):
            # Don't add the __init__ for this module (it IS this module)
            handled = {'__init__' : 1}
            path = module.__path__[0]
            for file in os.listdir(path):
                modname = inspect.getmodulename(file)
                if modname and not handled.get(modname):
                    handled[modname] = 1
                    contents.append((modname, name, 0))
                elif pydoc.ispackage(os.path.join(path, file)):
                    # 'file' will be a directory name
                    contents.append((file, name, 1))
            contents.sort()
            self.start_element('package-index')
            for name, path, ispackage in contents:
                tagname = ispackage and 'package' or 'module'
                attributes = {'name' : name,
                              'path' : path,
                              'public' : self.ispublic(name),
                              }
                self.write_element(tagname, attributes)
            self.end_element('package-index')

        modules = inspect.getmembers(module, inspect.ismodule)
        if modules:
            # Do not document modules covered by package-index
            # or are aliases for modules contained herein
            names = map(lambda x: x[1] + '.' + x[0], contents)
            modules = filter(lambda m, o=names, n=module.__name__:
                             m[1].__name__ not in o or
                             m[1].__name__ != n + '.' + m[0],
                             modules)
            self.section('modules', modules, self.doc_submodule)

        #classes = inspect.getmembers(module, self.isclass)
        classes = inspect.getmembers(module, inspect.isclass)
        if classes:
            self.section('classes', classes, self.doc_class)

        #funcs = inspect.getmembers(module, self.isroutine)
        funcs = inspect.getmembers(module, inspect.isroutine)
        if funcs:
            self.section('functions', funcs, self.doc_function)
            
        others = inspect.getmembers(module, pydoc.isdata)
        if others:
            others = filter(lambda (name, value):
                            name not in ['__builtins__', '__doc__', '__all__'],
                            others)
            self.section('others', others, self.doc_other)

        self.end_element(root_element)
        return

    def doc_submodule(self, module, name):
        """Produce XML documentation for a submodule"""
        realname = module.__name__
        name = name or realname

        attributes = {'name' : name,
                      'public' : self.ispublic(name),
                      'documented' : self.isdocumented(realname),
                      }
        if name != realname:
            attributes['realname'] = realname

        self.start_element('module', attributes)

        desc = self.escape(pydoc.getdoc(module))
        self.write_element('description', content=desc)

        self.end_element('module')
        return

    def doc_class(self, klass, name):
        """Produce XML documentation for a given class object."""
        realname = klass.__name__
        name = name or realname

        attributes = {'name' : name,
                      'public' : self.ispublic(name) and 'yes' or 'no',
                      }
        if name != realname:
            attributes['realname'] = realname

        self.start_element('class', attributes)

        for base in klass.__bases__:
            attributes = {'module' : base.__module__,
                          'name' : base.__name__,
                          'documented' : self.isdocumented(base.__module__),
                          }
            self.write_element('base', attributes)

        desc = self.escape(pydoc.getdoc(klass))
        self.write_element('description', content=desc)

        self.start_element('method-resolution-order')
        mro = list(inspect.getmro(klass))
        for base in mro:
            attributes = {'module' : base.__module__,
                          'name' : base.__name__,
                          }
            self.write_element('base', attributes)
        self.end_element('method-resolution-order')

        attrs = inspect.classify_class_attrs(klass)
        while attrs:
            if mro:
                thisclass = mro.pop(0)
            else:
                thisclass = attrs[0][2]

            inherited = [ t for t in attrs if t[2] is not thisclass ]
            attrs = [ t for t in attrs if t[2] is thisclass ]

            if thisclass is klass:
                inherited = ''
            else:
                inherited = pydoc.classname(thisclass, klass.__module__)

            # Sort attrs by name.
            attrs.sort(lambda t1, t2: cmp(t1[0], t2[0]))

            # This captures all method types
            methods = [ t for t in attrs if t[1].endswith('method') ]
            if methods:
                attributes = {}
                if inherited:
                    attributes['inherited'] = inherited
                self.start_element('methods', attributes)

                for name, kind, home_class, method in methods:
                    # inspect.classify_class_attrs gets the methods from the 
                    # class __dict__ which returns functions, not methods
                    method = getattr(klass, name)
                    self.doc_method(method, name, klass, kind)

                self.end_element('methods')

            # Python 2.2 only
            properties = [ t for t in attrs if t[1] == 'property' ]
            if properties:
                attributes = {}
                if inherited:
                    attributes['inherited'] = inherited
                self.start_element('properties', attributes)

                for name, kind, home_class, prop in properties:
                    self.doc_property(prop, name, klass)

                self.end_element('properties')

            data = [ t for t in attrs if t[1] == 'data' ]
            if data:
                attributes = {}
                if inherited:
                    attributes['inherited'] = inherited
                self.start_element('others', attributes)

                for name, kind, home_class, other in data:
                    # inspect.classify_class_attrs gets the attributes from the 
                    # class __dict__
                    other = getattr(klass, name)
                    self.doc_other(other, name)

                self.end_element('others')
            attrs = inherited

        self.end_element('class')
        return

    def format_arg(self, arg, default=None):
        attributes = {}
        if default is not None:
            attributes['default'] = default
        if type(arg) in [types.TupleType, types.ListType]:
            self.start_element('sequence', attributes)
            for a in arg:
                self.format_arg(a)
            self.end_element('sequence')
        else:
            attributes['name'] = arg
            self.write_element('arg', attributes)
        return

    def doc_args(self, object):
        if inspect.ismethod(object):
            object = object.im_func

        self.start_element('arguments')

        if inspect.isfunction(object):
            args, varargs, varkw, defaults = inspect.getargspec(object)

            if defaults:
                firstdefault = len(args) - len(defaults)

            for i in range(len(args)):
                if defaults and i >= firstdefault:
                    default = repr(defaults[i - firstdefault])
                else:
                    default = None
                self.format_arg(args[i], default)

            if varargs:
                self.write_element('var-args', {'name' : varargs})

            if varkw:
                self.write_element('var-keywords', {'name' : varkw})
        else:
            self.write_element('unknown')

        self.end_element('arguments')
        return

    def doc_method(self, method, name, klass, kind):
        """
        Document a method, class method or static method as given by 'kind'
        """
        realname = method.__name__
        name = name or realname
        attributes = {'name' : name,
                      'id' : klass.__name__ + '-' + name,
                      'public' : self.ispublic(name) and 'yes' or 'no',
                      }
        if name != realname:
            attributes['realname'] = realname
            # aliased methods fail the 'is' test but compare true
            if getattr(klass, realname, None) == method:
                attributes['realid'] = klass.__name__ + '-' + realname

        tagname = kind.replace(' ', '-')
        self.start_element(tagname, attributes)

        desc = self.escape(pydoc.getdoc(method))
        self.write_element('description', content=desc)

        self.doc_args(method)

        # In Python 2.2 it could also be a method descriptor which is
        # just a wrapper that acts like a method
        if inspect.ismethod(method):
            imclass = method.im_class
            if imclass is not klass:
                attributes = {'module' : imclass.__module__,
                              'name' : imclass.__name__,
                              }
                self.write_element('inherited-from', attributes)
            elif imclass:
                # This always returns the passed in class as the first item
                for base in inspect.getmro(imclass)[1:]:
                    overridden = getattr(base, name, None)
                    if overridden:
                        attributes = {'module' : base.__module__,
                                      'name' : base.__name__,
                                      }
                        self.write_element('overrides', attributes)
                        break

        self.end_element(tagname)
        return

    def doc_property(self, prop, name, klass):
        """
        Document a property (computed attribute).
        """
        attributes = {'name' : name,
                      'public' : self.ispublic(name) and 'yes' or 'no',
                      }
        self.start_element('property', attributes)

        desc = self.escape(pydoc.getdoc(prop))
        self.write_element('description', content=desc)

        for attr, tag in [('fget', 'getter'),
                          ('fset', 'setter'),
                          ('fdel', 'deleter')]:
            func = getattr(prop, attr)
            if inspect.isroutine(func):
                name = func.__name__
                attributes = {
                    'name' : name,
                    'id' : klass.__name__ + '-' + name,
                    }
                if getattr(klass, name, None) is func:
                    attributes['function-id'] = klass.__name__ + '-' + name
                self.write_element(tag, attributes)

        self.end_element('property')
        return

    def doc_function(self, func, name):
        """
        Document a function
        """
        realname = func.__name__
        if realname == '<lambda>':
            realname = 'lambda'
        name = name or realname

        attributes = {'name' : name,
                      'id' : name,
                      'public' : self.ispublic(name) and 'yes' or 'no',
                      }
        if name != realname:
            attributes['realname'] = realname

        self.start_element('function', attributes)

        desc = self.escape(pydoc.getdoc(func))
        self.write_element('description', content=desc)

        self.doc_args(func)

        self.end_element('function')
        return

    def doc_other(self, object, name):
        """Produce XML documentation for a data object."""
        attributes = {'name' : name,
                      'public' : self.ispublic(name) and 'yes' or 'no',
                      }
        self.start_element('other', attributes)

        # SWIG wrapped objects throw exceptions with getattr(obj, attr, None)
        # but not with hasattr(obj, attr)
        if hasattr(object, '__doc__') and getattr(object, '__doc__'):
            doc = self.escape(getattr(object, '__doc__'))
            self.write_element('description', content=doc)

        self.write_element('value', content=self.repr(object))

        self.end_element('other')
        return
