import os
from distutils import util
from distutils.command import install

INSTALL_SCHEMES = {}
for scheme, type_dict in install.INSTALL_SCHEMES.items():
    INSTALL_SCHEMES[scheme] = type_dict = type_dict.copy()

    # First we want ALL installs to go to the platlib directory
    type_dict['purelib'] = type_dict['platlib']

    # 4Suite needs all files to be under the Ft.Share module
    base = '$platlib/Ft/Share'
    type_dict['data'] = base
    type_dict['scripts'] = base + '/Bin'
    type_dict['docs'] = base + '/Docs'
    type_dict['tests'] = base + '/Tests'

EXTRA_KEYS = ('docs', 'tests')

SCHEME_KEYS = install.SCHEME_KEYS + EXTRA_KEYS

class Install(install.install):

    command_name = 'install'

    user_options = install.install.user_options + [
        ('install-docs=', None, 'installation directory for documentation'),
        ('install-tests=', None, 'installation directory for tests'),

        ('with-docs', None, 'enable documentation build'),
        ('without-docs', None, 'disable documentation build [default]'),
        ]

    boolean_options = install.install.boolean_options + \
                      ['with-docs', 'without-docs']

    negative_opt = install.install.negative_opt.copy()
    negative_opt['without-docs'] = 'with-docs'

    def initialize_options(self):
        install.install.initialize_options(self)
        self.install_docs = None
        self.install_tests = None
        self.with_docs = 0
        return

    def finalize_options(self):
        # Pass the documentation flag to the build command if
        # it hasn't already been run.
        build = self.distribution.get_command_obj('build')
        if not build.finalized:
            build.with_docs = self.with_docs
        
        install.install.finalize_options(self)
        apply(self.convert_paths, EXTRA_KEYS)
        if self.root is not None:
            apply(self.change_roots, EXTRA_KEYS)
        return

    def _expand_attrs(self, attrs):
        # Allow attrs to reference other attrs already defined
        config_vars = self.config_vars.copy()
        for attr in attrs:
            val = getattr(self, attr, None)
            if val is not None:
                if os.name == 'posix':
                    val = os.path.expanduser(val)
                val = util.subst_vars(val, config_vars)
                if attr.startswith('install_'):  # added
                    config_vars[attr[8:]] = val  # added
                setattr(self, attr, val)

    def expand_dirs (self):
        self._expand_attrs(['install_purelib',
                            'install_platlib',
                            'install_lib',
                            'install_headers',
                            'install_scripts',
                            'install_data',
                            'install_docs',
                            'install_tests',
                            ])

    def select_scheme(self, name):
        # it's the caller's problem if they supply a bad name!
        scheme = INSTALL_SCHEMES[name]
        for key in SCHEME_KEYS:
            attrname = 'install_' + key
            if getattr(self, attrname) is None:
                setattr(self, attrname, scheme[key])

    # -- Predicates for sub-command list -------------------------------

    def has_docs(self):
        return self.with_docs and self.distribution.has_docs()

    def has_tests(self):
        return self.distribution.has_tests()

    # a list of commands this command might have to run to do its work.
    sub_commands = install.install.sub_commands + [
        ('install_docs', has_docs),
        ('install_tests', has_tests),
        ]
