########################################################################
#
# File Name:            Export.py
#
# Documentation:        http://docs.4suite.org/4SS/Lib/CommandLine.py.html
#
"""
WWW: http://4suite.org        e-mail: support@4suite.org

Copyright (c) 2001 Fourthought Inc, USA.   All Rights Reserved.
See  http://4suite.org/COPYRIGHT  for license and copyright information
"""
import os, sys
from Ft.Server.Client import FtServerClientException
from Ft.Server.Client.Commands import CommandUtil
from Ft.Server.Common.Install import InstallUtil
from Ft.Server.Common.ResourceTypes import ResourceType
from Ft.Server.Common import AclConstants, ResourceTypes
from Ft.Lib import Uuid
from distutils import archive_util, dir_util

__doc__ = """This command will export an application from the repository and create a setup.xml file.  You will be propted for a user name and password.  You must have read access to all of the files in the directory."""


def Run(options, args):

    baseLocalDir = options.get('directory','.') + os.sep
    if not os.path.isdir(baseLocalDir):
        print "Local path %s is not a directory" % baseLocalDir
    baseRepoDir = args.get('path','/')

    firstDirectory = None
    try:
        repo = CommandUtil.GetRepository(options,'4ss.export')
        try:
            curRepo = repo.fetchResource(baseRepoDir)

            if curRepo.getAbsolutePath() == '/':
                commonPath = '/'
            else:
                parent = curRepo.getParent()
                commonPath = parent.getAbsolutePath()
                if not commonPath[-1] == '/': commonPath += ''

            packageDirName = curRepo.getName()
            if not curRepo.isResourceType(ResourceType.CONTAINER):
                print "%s must be a container" % baseRepoDir
                return
            resourceList,firstDirectory = ExportContainer(curRepo,baseLocalDir,commonPath)
        finally:
            try:
                repo.txRollback()
            except:
                pass
    except FtServerClientException, e:
        sys.stderr.write(str(e)+'\n')
        return

    if packageDirName == "/":
        setupPath = os.path.join(baseLocalDir, 'setup.xml')
    else:
        setupPath = os.path.join(baseLocalDir, packageDirName, 'setup.xml')
    refUri = os.path.join(baseLocalDir,packageDirName,'')
    print "Writing Setup File %s" % setupPath
    p = InstallUtil.Product(resourceList,
                            name=options.get('title'),
                            version=float(options.get('version','0')),
                            description=options.get('description'))
    f = open(setupPath,'w')
    p.serialize(f,refUri=refUri)
    f.close()

    #See if they want zip or tar
    if options.get('archive'):
        #Archive it!!
        base_name = options.get('title',None)
        if not base_name:
            base_name = resourceList[0].path[1:]
        if options.get('version'):
            base_name += '-' + options['version']

        print "Creating %s archive..." % options['archive']
        fName = archive_util.make_archive (base_name,
                                   options['archive'],
                                   root_dir=baseLocalDir,
                                   base_dir=resourceList[0].path[1:],
                                   verbose=0,
                                   dry_run=0)

        print "Created: %s" % fName
        dir_util.remove_tree(firstDirectory)
        

def ExportContainer(cont,localDir,commonPath):
    """
    Export a specific container in the repository.  Gather a list of all items that need to be added
    to the setup file and store them in setupInfo
    """
    rl = []

    
    path = cont.getAbsolutePath()
    if path != '/':
        path = path[len(commonPath):]
    
    rl.append(InstallUtil.Container(path, GetAcl(cont), GetOwner(cont),
                                    GetImt(cont), GetDocDef(cont, commonPath)))

    #Make it locally
    if path == '/':
        newLocalDir = localDir
    else:
        newLocalDir = os.path.join(localDir, cont.getName()) + os.sep
        os.mkdir(newLocalDir)

    #Now, do all of its children
    for child in cont:
        print "Exporting: %s" % child.getAbsolutePath()
        if child.isResourceType(ResourceType.CONTAINER):
            rl.extend(ExportContainer(child,newLocalDir, commonPath)[0])
        else:
            rl.extend(ExportRawFile(child,newLocalDir, commonPath))
    return rl,newLocalDir

def ExportRawFile(rf, localDir, commonPath):

    #Write the content
    localFileName = os.path.join(localDir, rf.getName())
    open(localFileName,'w').write(rf.getContent())

    acl = GetAcl(rf)
    owner = GetOwner(rf)
    imt = GetImt(rf)
    path = rf.getAbsolutePath()
    path = path[len(commonPath):]

    if rf.resourceType == ResourceTypes.ResourceType.XML_DOCUMENT:
        r = ExportXmlDocument(rf,path,acl,owner,imt,commonPath)
    elif rf.resourceType == ResourceTypes.ResourceType.RDF_DOCUMENT:
        r = ExportRdfDocument(rf,path,acl,owner,imt,commonPath)
    elif rf.resourceType == ResourceTypes.ResourceType.XSLT_DOCUMENT:
        r = ExportXsltDocument(rf,path,acl,owner,imt,commonPath)
    elif rf.resourceType == ResourceTypes.ResourceType.SCHEMATRON_DOCUMENT:
        r = ExportSchematronDocument(rf,path,acl,owner,imt,commonPath)
    elif rf.resourceType == ResourceTypes.ResourceType.XPATH_DOCUMENT_DEFINITION:
        r = ExportXPathDocumentDefinition(rf,path,acl,owner,imt,commonPath,localDir)
    elif rf.resourceType == ResourceTypes.ResourceType.XSLT_DOCUMENT_DEFINITION:
        r = ExportXsltDocumentDefinition(rf,path,acl,owner,imt,commonPath,localDir)
    elif rf.resourceType == ResourceTypes.ResourceType.USER:
        r = ExportUser(rf,path,acl,owner,imt,commonPath)
    elif rf.resourceType == ResourceTypes.ResourceType.GROUP:
        r = ExportGroup(rf,path,acl,owner,imt,commonPath)
    elif rf.resourceType == ResourceTypes.ResourceType.SERVER:
        r = ExportServer(rf,path,acl,owner,imt,commonPath)
    elif rf.resourceType == ResourceTypes.ResourceType.COMMAND:
        r = ExportCommand(rf,path,acl,owner,imt,commonPath)
    elif rf.resourceType == ResourceTypes.ResourceType.ALIAS:
        r = ExportAlias(rf,path,acl,owner,imt,commonPath)
    elif rf.resourceType == ResourceTypes.ResourceType.URI_REFERENCE_FILE:
        r = ExportUriReferenceFile(rf,path,acl,owner,imt,commonPath)
    elif rf.resourceType == ResourceTypes.ResourceType.RAW_FILE:
        r = InstallUtil.RawFile(path,acl,owner,imt)
    else:
        raise "Huh? %s" % str(rf.resourceType)

    if type(r) != type([]):
        r = [r]
    for x in r:
        x.setPath(localFileName)
    return r

def ExportXmlDocument(res,path,acl,owner,imt,commonPath):
    dd = GetDocDef(res,commonPath)
    return InstallUtil.XmlDocument(path,acl,owner,imt,dd)
def ExportRdfDocument(res,path,acl,owner,imt,commonPath):
    dd = GetDocDef(res,commonPath)
    return InstallUtil.RdfDocument(path,acl,owner,imt,dd)
def ExportXsltDocument(res,path,acl,owner,imt,commonPath):
    ddPath = GetDocDef(res,commonPath)
    return InstallUtil.XsltDocument(path,acl,owner,imt,ddPath)
def ExportXsltDocumentDefinition(res,path,acl,owner,imt,commonPath,localDir):
    dd = GetDocDef(res,commonPath)
    ext = ExportExtModules(res,localDir) 
    ext.append(InstallUtil.XsltDocumentDefinition(path,acl,owner,imt,dd))
    return ext
def ExportXPathDocumentDefinition(res,path,acl,owner,imt,commonPath,localDir):
    dd = GetDocDef(res,commonPath)
    ext = ExportExtModules(res,localDir) 
    ext.append(InstallUtil.XPathDocumentDefinition(path,acl,owner,imt,dd))
    return ext
def ExportServer(res,path,acl,owner,imt,commonPath):
    dd = GetDocDef(res,commonPath)
    return InstallUtil.Server(path,acl,owner,imt,dd)
def ExportCommand(res,path,acl,owner,imt,commonPath):
    dd = GetDocDef(res,commonPath)
    return InstallUtil.Command(path,acl,owner,imt,dd)

def ExportAlias(res,path,acl,owner,imt,commonPath):
    dd = GetDocDef(res,commonPath)
    return InstallUtil.Alias(path,acl,owner,imt,dd,res.getReference().getAbsolutePath())

def ExportUser(res,path,acl,owner,imt,commonPath):
    dd = GetDocDef(res,commonPath)
    return InstallUtil.User(os.path.basename(path),
                            acl,
                            owner,
                            imt,
                            dd,
                            "/".join(path.split("/")[:-1]),
                            None,
                            res.getUserData(),
                            res.getPassword(),
                            )

def ExportGroup(res,path,acl,owner,imt,commonPath):
    dd = GetDocDef(res,commonPath)
    members = map(lambda x:x.getUsername(),res.getMembers())
    return InstallUtil.Group(os.path.basename(path), acl, owner,
                            imt, dd, "/".join(path.split("/")[:-1]),
                            members
                            )


def ExportExtModules(docdef,localDir):
    res = []
    for m in docdef.getCreationParams().extModules:
        if str(m[:26]) == "Ft.Share.ExtensionModules.":
            print "Exporting: ", m
            res.append(InstallUtil.ExtensionModule(m[26:] + '.py'))
            localFile = os.path.join(localDir,res[-1]._file)
            srcFile = os.path.join(InstallUtil.EXT_DIR,res[-1]._file)
            open(localFile,'w').write(open(srcFile).read())
            
    return res

def GetAcl(res):
    aclDict = res.getAcl()
    acl = []
    for access,info in aclDict.items():
        for ident,allowed in info.items():
            acl.append((access,ident,allowed))
    return acl

def GetOwner(res):
    owner = res.getOwner()
    ownerName = owner and owner.getUsername() or AclConstants.WORLD_GROUP_NAME
    if ownerName == res.getRoot().getCurrentUser().getUsername():
        ownerName = None
    return ownerName

def GetImt(res):
    imt = res.getImt()
    if res.resourceType == ResourceTypes.ResourceType.RAW_FILE and imt == 'text/plain':
        imt = None
    if res.resourceType in ResourceTypes.XML_DOCUMENTS and imt == 'text/xml':
        imt = None
    return imt

def GetDocDef(res,commonPath):
    dd = res.getDocumentDefinition()
    if dd is not None:
        dd = dd.getAbsolutePath()
        if dd[:len(commonPath)] == commonPath:
            dd = dd[len(commonPath):]
    return dd


def Register():

    from Ft.Lib.CommandLine import Options, Command, Arguments

    cmd = Command.Command('export',
                          """export an application.""",
                          '/fourthought.com',
                          __doc__,
                          function = Run,
                          arguments = [Arguments.OptionalArgument('path',
                                                                  "The base directory of the application to export.  The default is /.",
                                                                  str),
                                       ],
                          options = Options.Options([Options.Option('d',
                                                                    'directory=',
                                                                    "Directory to export too."),
                                                     Options.Option(None,
                                                                    'version=',
                                                                    "Product Version."),
                                                     Options.Option(None,
                                                                    'description=',
                                                                    "Product Description."),
                                                     Options.Option(None,
                                                                    'title=',
                                                                    "Product Title."),
                                                     Options.TypedOption(None,
                                                                         'archive=',
                                                                         "Create an archive (not all options will work on all platforms!!)",
                                                                         map(lambda x: (x[0],x[1][-1]),archive_util.ARCHIVE_FORMATS.items()),
                                                                         ),
                                                     ]),

                          fileName = __file__,
                          )
    return cmd


