########################################################################
#
# File Name:            DocumentReferenceImp.py
#
# Documentation:        http://docs.4suite.org/4SS/DocumentReferenceImp.py.html
#
"""
An intelligent reference to a document within or without 4SS
WWW: http://4suite.org         e-mail: support@4suite.org

Copyright (c) 2000-2001 Fourthought, Inc. USA.   All Rights Reserved.
See  http://4suite.org/COPYRIGHT  for license and copyright information
"""

from Ft.Lib import UriException
from Ft.Lib.Uri import BASIC_RESOLVER
from Ft.Xml import XPath
from Ft.Xml.Xslt import StylesheetReader
from Ft.Server import FTSERVER_NAMESPACE
from Ft.Server.Server import FtServerServerException, Error
from Ft.Server.Common import XmlLib, ResourceTypes
from Ft.Server.Server.Drivers import FtssInputSource
#from Ft.Server.Common import ResourceTypes
import SchematronStylesheet

class DocumentReferenceType:
    INTERNAL = 1
    EXTERNAL = 2
    STRING = 3


class DocumentReference:
    """
    Abstract base class for the various types of Document References.
    """
    def __init__(self,rType):
        self.referenceType = rType

    def toDom(self,baseObject):
        """
        Overidden by subclasses for implementation to convert the reference into a DOM node
        """
        raise FtServerServerException(Error.INTERNAL_ERROR,message = "Must override toDom in %s" % str(self))
    def toStylesheet(self,baseObject):
        """
        Overidden by subclasses for implementation to convert the reference into a stylesheet
        """
        raise FtServerServerException(Error.INTERNAL_ERROR,message = "Must override toStylesheet in %s" % str(self))
    def toSchematron(self,baseObject):
        """
        Overidden by subclasses for implementation to convert the reference into a schematron instance
        """
        raise FtServerServerException(Error.INTERNAL_ERROR,message = "Must override toSchematron in %s" % str(self))


class StringDocumentReference(DocumentReference):
    """
    Reference to a string array
    """
    def __init__(self,data,baseUri):
        self.data = data
        self.baseUri = baseUri
        DocumentReference.__init__(self,DocumentReferenceType.STRING)

    def toDom(self,baseObject):
        return FtssInputSource.NonvalidatingReader.parseString(self.data,self.baseUri,baseObject._driver)


    def toStylesheet(self,baseObject):
        isrc = FtssInputSource.FtssInputSourceFactory.fromString(self.data,
                                                                 self.baseUri,
                                                                 baseObject._driver)
            
        reader = StylesheetReader.StylesheetReader()
        return reader.fromSrc(isrc)

    def toSchematron(self,baseObject):
        baseUri = self.baseUri or baseObject.getPath().absolutePath
        return SchematronStylesheet.ParseSchematron(self.data, baseUri,baseObject._driver)


class InternalDocumentReference(DocumentReference):
    """
    Reference to an internal resource
    """
    def __init__(self,uri):
        self.uri = uri
        DocumentReference.__init__(self,DocumentReferenceType.INTERNAL)

    def toDom(self,baseObject):
        return baseObject.fetchResource(self.uri).asDom()

    def toStylesheet(self,baseObject):
        return baseObject.fetchResource(self.uri).asStylesheet()

    def toSchematron(self,baseObject):
        return baseObject.fetchResource(self.uri).asSchematron()

class ExternalDocumentReference(DocumentReference):
    """
    Reference to an external resource
    """
    def __init__(self,uri):
        self.uri = uri
        DocumentReference.__init__(self,DocumentReferenceType.EXTERNAL)


    def toDom(self,baseObject):
        try:
            src = BASIC_RESOLVER.resolve(self.uri)
        except UriException:
            raise FtServerServerException(Error.UNKNOWN_URI, uri=self.uri)

        return FtssInputSource.NonvalidatingReader.parseString(
            src.read(), self.uri, baseObject._driver)

    def toStylesheet(self,baseObject):
        try:
            src = BASIC_RESOLVER.resolve(self.uri)
        except UriException:
            raise FtServerServerException(Error.UNKNOWN_URI, uri=self.uri)

        isrc = FtssInputSource.FtssInputSourceFactory.fromString(
            src.read(), self.uri, baseObject._driver)
        reader = StylesheetReader.StylesheetReader()
        return reader.fromSrc(isrc)


    def toSchematron(self,baseObject):
        try:
            src = BASIC_RESOLVER.resolve(self.uri)
        except UriException:
            raise FtServerServerException(Error.UNKNOWN_URI, uri=self.uri)

        return SchematronStylesheet.ParseSchematron(src.read(), self.uri,
                                                    baseObject._driver)



def _Serialize(doc,dr):
    ref = doc.createElementNS(FTSERVER_NAMESPACE, 'ftss:DocumentReference')
    if dr.referenceType == DocumentReferenceType.STRING:
        ref.setAttributeNS(None, 'type', 'STRING')
        ref.setAttributeNS(None, 'baseUri', dr.baseUri)
        ref.appendChild(doc.createTextNode(dr.data))
    elif dr.referenceType == DocumentReferenceType.INTERNAL:
        ref.setAttributeNS(None, 'type', 'INTERNAL')
        ref.appendChild(doc.createTextNode(dr.uri))
    elif dr.referenceType == DocumentReferenceType.EXTERNAL:
        ref.setAttributeNS(None, 'type', 'EXTERNAL')
        ref.appendChild(doc.createTextNode(dr.uri))
    return ref


_documentReferenceTypeExpression = XPath.Compile('string(@type)')
_documentReferenceTextChildExpression = XPath.Compile('string(text())')
_documentReferenceStringBaseExpression = XPath.Compile('string(@baseUri)')
def _Deserialize(con):
    val_doc_type = _documentReferenceTypeExpression.evaluate(con)
    val_doc_type = val_doc_type or 'STRING'
    if val_doc_type == 'INTERNAL':
        text = _documentReferenceTextChildExpression.evaluate(con)
        docRef = InternalDocumentReference(text)
    elif val_doc_type == 'EXTERNAL':
        text = _documentReferenceTextChildExpression.evaluate(con)
        docRef = ExternalDocumentReference(text)
    else:
        stringBase = _documentReferenceStringBaseExpression.evaluate(con)
        text = _documentReferenceTextChildExpression.evaluate(con)
        docRef = StringDocumentReference(text,stringBase)
    return docRef
