########################################################################
#
# File Name:   CoreFunctions.py
#
# Docs:        http://docs.4suite.org/XPATH/CoreFunctions.py.html
#
"""
The implementation of all of the core functions for the XPath spec.
WWW: http://4suite.org/XPATH        e-mail: support@4suite.org

Copyright (c) 2000-2001 Fourthought Inc, USA.   All Rights Reserved.
See  http://4suite.org/COPYRIGHT  for license and copyright information
"""

import cStringIO, types

from xml.dom import Node
from Ft.Xml import EMPTY_NAMESPACE
from Ft.Lib import number, boolean
from Ft.Xml.XPath import NamespaceNode
from Ft.Xml.XPath import NaN, Inf
from Ft.Xml.XPath import Util, Conversions
from Ft.Xml.XPath import NAMESPACE_NODE
from Ft.Xml.XPath import CompiletimeException, RuntimeException

try:
    import os, gettext
    locale_dir = os.path.split(__file__)[0]
    gettext.install('4Suite', locale_dir)
#except ImportError, IOError:
#Note, 1.5.2 has gettext, but no install
except (ImportError, AttributeError, IOError):
    def _(msg):
        return msg


class Types:
    NumberType = 0
    StringType = 1
    BooleanType = 2
    NodeSetType = 3
    ObjectType = 4

### Node Set Functions ###

def Last(context):
    """Function: <number> last()"""
    return context.size


def Position(context):
    """Function: <number> position()"""
    return context.position


def Count(context, nodeSet):
    """Function: <number> count(<node-set>)"""
    if type(nodeSet) != type([]):
        raise RuntimeException(RuntimeException.WRONG_ARGUMENTS, 'count', _("expected node set argument"))
    return len(nodeSet)


def Id(context, object):
    """Function: <node-set> id(<object>)"""
    import warnings
    warnings.warn("id() function not supported")
    #We do not (cannot, really) support the id() function
    return []
    id_list = []
    if type(object) != type([]):
        st = Conversions.StringValue(object)
        id_list = st.split()
    else:
        for n in object:
            id_list.append(Conversions.StringValue(n))

    doc = context.node.rootNode
    getElementById = getattr(doc, 'getElementById', None)
    if not getElementById:
        getElementById = lambda id, d=doc, f=Util.GetElementById: f(d, id)

    nodeset = []
    for id in id_list:
        try:
            element = getElementById(id)
            if element:
                nodeset.append(element)
        except ValueError:
            # This exception will only occur from Util.GetElementById
            raise RuntimeException(RuntimeException.WRONG_ARGUMENTS, 'id',
                                   _("argument not unique"))
    return nodeset


def LocalName(context, nodeSet=None):
    """Function: <string> local-name(<node-set>?)"""
    if nodeSet is None:
        node = context.node
    else:
        if type(nodeSet) != type([]):
            raise RuntimeException(RuntimeException.WRONG_ARGUMENTS, 'local-name', _("expected node set"))
        nodeSet = Util.SortDocOrder(context, nodeSet)
        if type(nodeSet) != type([]) or len(nodeSet) == 0:
            return u''
        node = nodeSet[0]

    node_type = getattr(node, 'nodeType', None)
    if node_type in (Node.ELEMENT_NODE, Node.ATTRIBUTE_NODE):
        # localName could be null
        return node.localName or u''
    elif node_type == NAMESPACE_NODE:
        # localName could be null
        return node.localName or u''
    elif node_type == Node.PROCESSING_INSTRUCTION_NODE:
        # target cannot be null
        return node.target
    return u''


def NamespaceUri(context, nodeSet=None):
    """Function: <string> namespace-uri(<node-set>?)"""
    if nodeSet is None:
        node = context.node
    else:
        if type(nodeSet) != type([]):
            raise RuntimeException(RuntimeException.WRONG_ARGUMENTS, 'namespace-uri', _("expected node set"))
        nodeSet = Util.SortDocOrder(context, nodeSet)
        if type(nodeSet) != type([]) or len(nodeSet) == 0:
            return u''
        node = nodeSet[0]

    # only elements and attributes have a namespace-uri
    node_type = getattr(node, 'nodeType', None)
    if node_type in (Node.ELEMENT_NODE, Node.ATTRIBUTE_NODE):
        return node.namespaceURI or u''
    return u''


def Name(context, nodeSet=None):
    """Function: <string> name(<node-set>?)"""
    if nodeSet is None:
        node = context.node
    else:
        if type(nodeSet) != type([]):
            raise RuntimeException(RuntimeException.WRONG_ARGUMENTS, 'name', _("expected node set"))
        nodeSet = Util.SortDocOrder(context, nodeSet)
        if type(nodeSet) != type([]) or len(nodeSet) == 0:
            return u''
        node = nodeSet[0]

    node_type = getattr(node, 'nodeType', None)
    if node_type in (Node.ELEMENT_NODE, Node.ATTRIBUTE_NODE):
        return node.nodeName
    elif node_type == NAMESPACE_NODE:
        # localName could be null
        return node.localName or u''
    elif node_type == Node.PROCESSING_INSTRUCTION_NODE:
        # target cannot be null
        return node.target
    return u''


### String Functions ###

def String(context, object=None):
    """Function: <string> string(<object>?)"""
    if type(object) is types.UnicodeType:
        return object
    if object is None:
        object = context.node
    return Conversions.StringValue(object)


def Concat(context, *args):
    """Function: <string> concat(<string>, <string>, ...)"""
    if len(args) < 1:
        raise RuntimeException(RuntimeException.WRONG_ARGUMENTS, 'concat', _("at least 2 arguments expected"))
    return reduce(lambda a, b: a + Conversions.StringValue(b), args, '')


def StartsWith(context, outer, inner):
    """Function: <string> starts-with(<string>, <string>)"""
    outer = Conversions.StringValue(outer)
    inner = Conversions.StringValue(inner)
    return outer[:len(inner)] == inner and boolean.true or boolean.false


def Contains(context, outer, inner):
    """Function: <string> contains(<string>, <string>)"""
    if type(outer) is not types.UnicodeType:
        outer = Conversions.StringValue(outer)
    if type(inner) is not types.UnicodeType:
        inner = Conversions.StringValue(inner)
    return outer.find(inner) >= 0 and boolean.true or boolean.false


def SubstringBefore(context, outer, inner):
    """Function: <string> substring-before(<string>, <string>)"""
    if type(outer) is not types.UnicodeType:
        outer = Conversions.StringValue(outer)
    if type(inner) is not types.UnicodeType:
        inner = Conversions.StringValue(inner)
    index = outer.find(inner)
    if index == -1:
        return u''
    return outer[:index]


def SubstringAfter(context, outer, inner):
    """Function: <string> substring-after(<string>, <string>)"""
    if type(outer) is not types.UnicodeType:
        outer = Conversions.StringValue(outer)
    if type(inner) is not types.UnicodeType:
        inner = Conversions.StringValue(inner)
    index = outer.find(inner)
    if index == -1:
        return u''
    return outer[index+len(inner):]


def Substring(context, string, start, length=None):
    """Function: <string> substring(<string>, <number>, <number>?)"""
    if type(string) is not types.UnicodeType:
        string = Conversions.StringValue(string)
    if type(start) is not types.FloatType:
        start = Conversions.NumberValue(start)

    # NaN and +Inf indicate no substring
    if number.isnan(start) or (number.isinf(start) and start > 0):
        return u''

    if start < 1:
        # This handles -Inf which explodes int()
        start = 0
    else:
        # start is guaranteed to be at least 1
        # convert to python (0-based) index
        start = int(round(start)) - 1

    if length is None:
        return string[start:]
    elif type(length) is not types.FloatType:
        length = Conversions.NumberValue(length)
    if number.isnan(length):
        return u''
    length = int(round(length))
    if length <= 0:
        return u''
    return string[start:start + length]


def StringLength(context, st=None):
    """Function: <number> string-length(<string>?)"""
    if st is None:
        st = context.node
    return len(Conversions.StringValue(st))


def Normalize(context, st=None):
    """Function: <string> normalize-space(<string>?)"""
    if st is None:
        st = context.node
    st = Conversions.StringValue(st)
    return ' '.join(st.split())


def Translate(context, source, fromChars, toChars):
    """Function: <string> translate(<string>, <string>, <string>)"""
    source = Conversions.StringValue(source)
    fromChars = Conversions.StringValue(fromChars)
    # remove duplicate chars from From string
    fromChars = reduce(lambda st, c: st + c * (st.find(c) == -1), fromChars, '')
    toChars = Conversions.StringValue(toChars)[:len(fromChars)]

    # string.maketrans/translate do not handle unicode
    translate = {}
    for from_char, to_char in map(None, fromChars, toChars):
        translate[ord(from_char)] = to_char

    result = reduce(lambda a, b, t=translate:
                    a + (t.get(ord(b), b) or ''),
                    source, '')
    return result

### Boolean Functions ###

def Boolean(context, object):
    """Function: <boolean> boolean(<object>)"""
    return Conversions.BooleanValue(object)


def Not(context, object):
    """Function: <boolean> not(<boolean>)"""
    return (not Conversions.BooleanValue(object) and boolean.true) or boolean.false


def True(context):
    """Function: <boolean> true()"""
    return boolean.true


def False(context):
    """Function: <boolean> false()"""
    return boolean.false


def Lang(context, lang):
    """Function: <boolean> lang(<string>)"""
    lang = Conversions.StringValue(lang).upper()
    node = context.node
    while node:
        lang_attr = filter(lambda x:x.name == 'xml:lang' and x.value, node.attributes.values())
        value = lang_attr and lang_attr[0].nodeValue or None
        if value:
            # See if there is a suffix
            index = value.find('-')
            if index != -1:
                value = value[:index]
            value = value.upper()
            return value == lang and boolean.true or boolean.false
        node = node.nodeType == Node.ATTRIBUTE_NODE and node.ownerElement or node.parentNode
        if node.nodeType == Node.DOCUMENT_NODE:
            break
    return boolean.false

### Number Functions ###

def Number(context, object=None):
    """Function: <number> number(<object>?)"""
    if object is None:
        object = [context.node]
    return Conversions.NumberValue(object)


def Sum(context, nodeSet):
    """Function: <number> sum(<node-set>)"""
    nns = map(Conversions.NumberValue, nodeSet)
    return reduce(lambda x, y: x + y, nns, 0)


def Floor(context, object):
    """Function: <number> floor(<number>)"""
    num = Conversions.NumberValue(object)
    if number.isnan(num) or number.isinf(num):
        return num
    elif int(num) == num:
        return num
    elif num < 0:
        return int(num) - 1
    else:
        return int(num)


def Ceiling(context, object):
    """Function: <number> ceiling(<number>)"""
    num = Conversions.NumberValue(object)
    if number.isnan(num) or number.isinf(num):
        return num
    elif int(num) == num:
        return num
    elif num > 0:
        return int(num) + 1
    else:
        return int(num)


def Round(context, object):
    """Function: <number> round(<number>)"""
    num = Conversions.NumberValue(object)
    if number.isnan(num) or number.isinf(num):
        return num
    else:
        return round(num, 0)


### Function Mappings ###

CoreFunctions = {
    (EMPTY_NAMESPACE, 'last'): Last,
    (EMPTY_NAMESPACE, 'position'): Position,
    (EMPTY_NAMESPACE, 'count'): Count,
    (EMPTY_NAMESPACE, 'id'): Id,
    (EMPTY_NAMESPACE, 'local-name'): LocalName,
    (EMPTY_NAMESPACE, 'namespace-uri'): NamespaceUri,
    (EMPTY_NAMESPACE, 'name'): Name,
    (EMPTY_NAMESPACE, 'string'): String,
    (EMPTY_NAMESPACE, 'concat'): Concat,
    (EMPTY_NAMESPACE, 'starts-with'): StartsWith,
    (EMPTY_NAMESPACE, 'contains'): Contains,
    (EMPTY_NAMESPACE, 'substring-before'): SubstringBefore,
    (EMPTY_NAMESPACE, 'substring-after'): SubstringAfter,
    (EMPTY_NAMESPACE, 'substring'): Substring,
    (EMPTY_NAMESPACE, 'string-length'): StringLength,
    (EMPTY_NAMESPACE, 'normalize-space'): Normalize,
    (EMPTY_NAMESPACE, 'translate'): Translate,
    (EMPTY_NAMESPACE, 'boolean'): Boolean,
    (EMPTY_NAMESPACE, 'not'): Not,
    (EMPTY_NAMESPACE, 'true'): True,
    (EMPTY_NAMESPACE, 'false'): False,
    (EMPTY_NAMESPACE, 'lang'): Lang,
    (EMPTY_NAMESPACE, 'number'): Number,
    (EMPTY_NAMESPACE, 'sum'): Sum,
    (EMPTY_NAMESPACE, 'floor'): Floor,
    (EMPTY_NAMESPACE, 'ceiling'): Ceiling,
    (EMPTY_NAMESPACE, 'round'): Round,
    }

Args = {
    Substring : (Types.StringType, [Types.StringType, Types.StringType]),
    }
