########################################################################
# $Header: /var/local/cvsroot/4Suite/Ft/Xml/Xslt/Processor.py,v 1.51 2003/05/21 01:03:28 mbrown Exp $
"""
XSLT processing engine

Copyright 2003 Fourthought, Inc. (USA).
Detailed license and copyright information: http://4suite.org/COPYRIGHT
Project home, documentation, distributions: http://4suite.org/
"""

import os, sys, types, operator, cStringIO, warnings
from xml.dom import Node

from Ft import STRING_TYPES, DEFAULT_ENCODING
from Ft.Xml import InputSource, EMPTY_NAMESPACE
from Ft.Xml.Lib import StripElements
from Ft.Xml.XPath import Util
from Ft.Xml.Xslt import XsltContext, Error, XsltException, MessageSource
from Ft.Xml.Xslt import OutputHandler, RtfWriter, StringWriter
from Ft.Xml.Xslt import StylesheetReader
from Ft.Xml import Domlette
from Ft.Xml import InputSource
from Ft.Lib import Uri

# Media types that signal that an xml-stylesheet PI points to an XSLT
# document, when the PI contains a type pseudo-attribute.
#
# Note: RFC 3023 suggests application/xslt+xml, and says the +xml
# suffix is not required (but is a SHOULD). If users want to use the
# 'text/xsl' convention, do Processor.XSLT_IMT.append('text/xsl')
# after import but before instantiating Processor.Processor.
#
XSLT_IMT = ['application/xslt+xml', 'application/xslt',
            'text/xml', 'application/xml']

# For builtin extension elements/functions
import Exslt, BuiltInExtElements

class Processor:
    """
    An XSLT processing engine (4XSLT).

    Typical usage:

    from Ft.Lib.Uri import OsPathToUri
    from Ft.Xml import InputSource
    from Ft.Xml.Xslt import Processor
    # this is just one of several ways to create InputSources
    styuri = OsPathToUri('/absolute/path/to/stylesheet.xslt')
    srcuri = OsPathToUri('/absolute/path/to/doc.xml')
    STY = InputSource.DefaultFactory.fromUri(styuri)
    SRC = InputSource.DefaultFactory.fromUri(srcuri)
    proc = Processor.Processor()
    proc.appendStylesheet(STY)
    result = Processor.run(SRC)

    Optional constructor arguments are:

      stylesheetIncPaths: a list of alternate base URIs to use
        when resolving relative hrefs in xsl:import/xsl:include
        instructions (useful when hiding included/imported
        stylesheets in private directories).
        Can also be supplied via the XSLTINCLUDE environment
        variable. These base URIs are only tried when the
        standard XSLT behavior of using the base URI of the
        xsl:import/include element fails to result in
        retrieval of a document.

      useEnvironment: a flag indicating whether the processor
        should use or ignore environment variables that affect
        its configuration

      documentReader: an object that will be used to parse XML
        source documents (not stylesheets); defaults to
        Ft.Xml.Domlette.NonvalidatingReader. It can be any
        object that has a parse() method that returns a DOM or
        Domlette tree.

      implementation: the DOM implementation instance; necessary
        if supplying a documentReader. Defaults to
        Ft.Xml.Domlette.implementation. Needs to have a
        createRootNode() method.

    See the run() method for additional options.
    """
    _4xslt_trace = 0
    _4xslt_profile = 0
    _4xslt_debug = 0

    def __init__(self, stylesheetIncPaths=None, useEnvironment=1,
                 documentReader=None, implementation=None):
        self._suppressMsgs = 0
        self.stylesheet = None
        self.inputSourceFactory = InputSource.DefaultFactory
        self._stylesheetIncPaths = stylesheetIncPaths or []
        if useEnvironment:
            if os.environ.has_key('XSLTINCLUDE'):
                style_inc_path_list = [
                    Uri.OsPathToUri(path, attemptAbsolute=1)
                    for path in os.environ["XSLTINCLUDE"].split(os.pathsep)
                    ]
                self._stylesheetIncPaths.extend(style_inc_path_list)

        #Create the default reader for documents
        self._docReader = documentReader or Domlette.NonvalidatingReader

        self._domimp = implementation or Domlette.implementation

        self._lastOutputParams = None

        # has the "built-in template invoked with params" warning been issued?
        self._builtInWarningGiven = 0

        self.extFunctions = {}  #Cache ext functions to give to the context

        self.extElements = {}
        self.extElements.update(Exslt.ExtElements)
        self.extElements.update(BuiltInExtElements.ExtElements)

        # Create a single instance of this to be used every where
        self._styReader = StylesheetReader.StylesheetReader(
            self._stylesheetIncPaths
            )
        self.extensionParams = {}
        return

    def getStripElements(self):
        if self.stylesheet:
            return self.stylesheet.spaceRules
        else:
            return []

    def registerExtensionModules(self, moduleList):
        """
        Register a list of modules that support the public ExtFunction and/or
        ExtElement dictionaries
        """
        for mod_name in moduleList:
            if mod_name:
                mod = __import__(mod_name,{},{},['ExtFunctions'])
                if hasattr(mod,'ExtFunctions'):
                    self.extFunctions.update(mod.ExtFunctions)
                if hasattr(mod,'ExtElements'):
                    self.extElements.update(mod.ExtElements)
        return

    def registerExtensionFunction(self, namespace, localName, function):
        """
        Register a single extension function
        """
        self.extFunctions[(namespace, localName)] = function
        return

    def registerExtensionElement(self, namespace, localName, klass):
        """
        Register a single extension element
        """
        self.extElements[(namespace, localName)] = klass
        return

    def setDocumentReader(self, docReader):
        """
        Set the reader used to read in an input source.
        Commonly used to set the reader to Domlette.ValidatingReader, but anything with a
        parse interface that returns a DOM tree will work
        """
        self._docReader = docReader
        return

    def setStylesheetReader(self, reader):
        self._styReader = reader
        return

    def appendStylesheet(self, iSrc):
        """
        Append a stylesheet from an input source
        """
        sty = self._styReader.fromSrc(iSrc, extElements=self.extElements)
        if self.stylesheet:
            for child in self.stylesheet.children:
                child.importIndex += 1000
            for child in sty.children:
                self.stylesheet.appendChild(child)
                child.parent = self.stylesheet
            self.stylesheet.reset()
            self.stylesheet.setup()
        else:
            self.stylesheet = sty
        self.outputParams = self.stylesheet.outputParams
        return

    def appendStylesheetInstance(self, sty, refUri=''):
        '''
        Accepts a valid StyleDOM node.  We need a refUri because instant stylesheets
        can be moved around, therefore changing how imports and includes work
        '''
        if self.stylesheet:
            sty = self._styReader.fromInstant(sty, refUri, is_import=1)
            for child in self.stylesheet.children:
                child.importIndex += 1000
            for child in sty.children:
                self.stylesheet.appendChild(child)
                child.parent = self.stylesheet
            self.stylesheet.reset()
            self.stylesheet.setup()
        else:
            self.stylesheet = sty
        self.outputParams = self.stylesheet.outputParams
        return

    def appendStylesheetNode(self, node, refUri=''):
        """
        Append a stylesheet from a regular DOM node
        """
        sty = self._styReader.fromDocument(node, refUri)
        if self.stylesheet:
            for child in self.stylesheet.children:
                child.importIndex += 1000
            for child in sty.children:
                self.stylesheet.appendChild(child)
                child.parent = self.stylesheet
            self.stylesheet.reset()
            self.stylesheet.setup()
        else:
            self.stylesheet = sty
        self.outputParams = self.stylesheet.outputParams
        return

    def run(self, iSrc, ignorePis=0, topLevelParams=None,
            writer=None, outputStream=None):
        """
        Transform a source document as given via an InputSource.

        Assumes the Processor instance has already had stylesheets
        appended via appendStylesheet, or the source document contains
        xml-stylesheet processing instructions.

        The ignorePis flag, if set, will ignore xml-stylesheet PIs in
        the source document. The topLevelParams argument is an optional
        dictionary of stylesheet parameters, the keys of which may be
        given as strings if they have no namespace, or as
        (uri, localname) tuples otherwise. The optional writer argument
        is a SAX-like event handler that is an Ft.Xml.Xslt.NullWriter
        subclass. The default writer is either an Ft.Xml.Xslt.XmlWriter,
        HtmlWriter or PlainTextWriter, depending on the stylesheet(s).
        The optional outputStream argument is a Python file-like object
        to be used as the destination for the writer's output.
        """
        #Update the strip elements
        #Assume that the ones from XSLT have higher priority
        ns = self.getStripElements()
        for s in iSrc.stripElements:
            ns.append(s)
        iSrc.stripElements = ns
        self._documentInputSource = iSrc
        try:
            src = self._docReader.parse(iSrc)
        except Exception, e:
            raise XsltException(Error.SOURCE_PARSE_ERROR, iSrc.uri or '<Python string>', e)
        if not ignorePis and self.checkStylesheetPis(src, iSrc):
            #Do it again with updates WS strip lists

            #NOTE:  There is a case where this will produce the wrong results.  If, there were
            #previous stylesheets that defined removing white space, then the
            #processing instruction referenced a stylesheet that overrode some of these
            #whitespace processing rules, the original trimmed space will be lost

            #Regardless, we need to remove any new whitespace defined in the PI
            self._stripElements(src)

        return self.execute(src, iSrc, ignorePis, topLevelParams,
                            writer, outputStream)

    def runNode(self, node, sourceUri, ignorePis=0, topLevelParams=None,
                writer=None, outputStream=None,preserveSrc=0):
        """
        Transform a source document as given via a DOM Node.

        Assumes the Processor instance has already had stylesheets
        appended via appendStylesheet, or the source document contains
        xml-stylesheet processing instructions.

        The ignorePis flag, if set, will ignore xml-stylesheet PIs in
        the source document. The topLevelParams argument is an optional
        dictionary of stylesheet parameters, the keys of which may be
        given as strings if they have no namespace, or as
        (uri, localname) tuples otherwise. The optional writer argument
        is a SAX-like event handler that is an Ft.Xml.Xslt.NullWriter
        subclass. The default writer is either an Ft.Xml.Xslt.XmlWriter,
        HtmlWriter or PlainTextWriter, depending on the stylesheet(s).
        The optional outputStream argument is a Python file-like object
        to be used as the destination for the writer's output.

        preserveSrc signals that the source DOM should not be mutated.
        """

        if node.nodeType != Node.DOCUMENT_NODE:
            raise ValueError("Node must be a document")

        if preserveSrc:
            newDoc = self._domimp.createRootNode(sourceUri)
            for child in node.childNodes:
                new_node = newDoc.importNode(child,1)
                newDoc.appendChild(new_node)
                node = newDoc

        self._stripElements(node)

        #Create a dummy iSrc
        docInputSource = InputSource.InputSource(None,
                                                 sourceUri,
                                                 processIncludes=1,
                                                 stripElements=self.getStripElements(),
                                                 factory=self.inputSourceFactory)


        if not ignorePis and self.checkStylesheetPis(node, docInputSource):
            #Do it again with updates WS strip lists

            #NOTE:  There is a case where this will produce the wrong results.  If, there were
            #previous stylesheets that defined removing white space, then the
            #processing instruction referenced a stylesheet that overrode some of these
            #whitespace processing rules, the original trimmed space will be lost

            #Regardless, we need to remove any new whitespace defined in the PI
            self._stripElements(node)


        return self.execute(node,
                            docInputSource,
                            ignorePis=ignorePis,
                            topLevelParams=topLevelParams,
                            writer=writer,
                            outputStream=outputStream)

    def checkStylesheetPis(self, node, inputSource):
        #Note: A Stylesheet PI can only be in the prolog, acc to the NOTE
        #http://www.w3.org/TR/xml-stylesheet/

        root = node.rootNode

        hrefs = []
        for child in root.childNodes:
            if child.nodeType == Node.PROCESSING_INSTRUCTION_NODE:
                if child.target == 'xml-stylesheet':
                    data = child.data.split()
                    pseudo_attrs = {}
                    for d in data:
                        seg = d.split('=')
                        if len(seg) == 2:
                            pseudo_attrs[seg[0]] = seg[1][1:-1]

                    if pseudo_attrs.has_key('href'):
                        imt = pseudo_attrs.get('type')
                        if imt is None or imt in XSLT_IMT:
                            hrefs.append(pseudo_attrs['href'])

        if hrefs:
            self._styReader._import_index += 1
            for href in hrefs:
                # Resolve the PI from the document input Source
                new_source = inputSource.resolve(href, '', 'xml-stylesheet PI')
                self.appendStylesheet(new_source)

        # Return true if xml-stylesheet processing instructions are handled
        return not not hrefs

    def execute(self, node, docInputSource, ignorePis=0, topLevelParams=None,
                writer=None, outputStream=None):
        '''
        Warning: do not call this method directly unless you know what
        you're doing.  If unsure, you probably want the runNode method.

        Run the stylesheet processor against the given XML DOM node with the
        stylesheets that have been registered.  It does not mutate the source.
        If writer is given, it is used in place of the default output method
        decisions for choosing the proper writer.
        '''
        #FIXME: What about ws stripping?
        #ANSWER: Whitespace stripping happens only in the run*() interfaces.
        #  This method is use-at-your-own-risk. The XSLT conformance of the
        #  source is maintained by the caller. This exists as a performance
        #  hook.
        topLevelParams = topLevelParams or {}

        self.attributeSets = {}
        self.keys = {}

        #See f:chain-to extension element
        self.chainTo = None

        if not self.stylesheet:
            raise XsltException(Error.NO_STYLESHEET)

        # Use an internal stream to gather the output only if the caller
        # didn't supply other meanings of retrieving it.
        internalStream = writer is None and outputStream is None

        if not writer:
            # Use OutputHandler to determine the real writer to use.
            outputStream = outputStream or cStringIO.StringIO()
            writer = OutputHandler.OutputHandler(self.outputParams,
                                                 outputStream,
                                                 self.__writerChanged)
        self.writers = [writer]

        # Setup the named templates
        self._namedTemplates = self.stylesheet.getNamedTemplates()

        # Initialize any stylesheet parameters
        tlp = topLevelParams.copy()
        self._normalizeParams(tlp)
        self._docInputSource = docInputSource

        # Prepare the stylesheet for processing
        self.stylesheet.primeStylesheet(node, self, tlp, docInputSource.uri)

        # Create the context used for processing
        variables = self.stylesheet.getGlobalVariables()
        functions = self.stylesheet.getInitialFunctions()
        context = XsltContext.XsltContext(node, 1, 1, None,
                                          varBindings=variables,
                                          processor=self,
                                          extFunctionMap=functions)
        context.documents.update(self.stylesheet.root.sourceNodes)
        context.addDocument(node, docInputSource.uri)

        # Process the document
        self.writers[-1].startDocument()
        self.applyTemplates(context)
        self.writers[-1].endDocument()

        # Perform cleanup
        self.stylesheet.idle(node, self, docInputSource.uri)
        Util.FreeDocumentIndex(node)

        #How does this contrast with access to self.outputParams ?
        self._lastOutputParams = self.writers[-1]._outputParams

        if internalStream:
            # Get the result from our cStringIO 'stream'.
            result = outputStream.getvalue()
        else:
            # It is the callers responsibility to get the result
            result = u""
        return result

    def applyTemplates(self, context, params=None):
        params = params or {}
        if not self.stylesheet.applyTemplates(context, self, params):
            # No matching templates found, use builtin templates
            if params and not self._builtInWarningGiven:
                self.warning(MessageSource.BUILTIN_TEMPLATE_WITH_PARAMS)
                self._builtInWarningGiven = 1
            if context.node.nodeType == Node.TEXT_NODE:
                self.writers[-1].text(context.node.data)
            elif context.node.nodeType in [Node.ELEMENT_NODE, Node.DOCUMENT_NODE]:
                state = context.copy()
                node_set = context.node.childNodes
                size = len(node_set)
                pos = 1
                for node in node_set:
                    context.node, context.position, context.size = \
                                  node, pos, size
                    self.applyTemplates(context)
                    pos += 1
                context.set(state)
            elif context.node.nodeType == Node.ATTRIBUTE_NODE:
                self.writers[-1].text(context.node.value)
        return

    # FIXME: l10n
    def xslMessage(self, msg):
        if not self._suppressMsgs:
            sys.stderr.write("STYLESHEET MESSAGE:\n")
            sys.stderr.write(msg)
            sys.stderr.write("\nEND STYLESHEET MESSAGE\n")
        return

    def warning(self, message):
        if not self._suppressMsgs:
            # Using level=2 to show the stack where the warning occured.
            warnings.warn(message, stacklevel=2)
        return

    def messageControl(self, suppress):
        self._suppressMsgs = suppress
        return

    def __writerChanged(self, newWriter):
        """
        Private callback function to allow the OutputHandler to remove
        itself from the writer stack.
        """
        self.writers[-1] = newWriter
        return

    def addHandler(self, outputParams, stream):
        """
        Add the handler used to determine real writer used for a particular
        output method.
        """
        handler = OutputHandler.OutputHandler(outputParams, stream,
                                              self.__writerChanged)
        self.writers.append(handler)
        handler.startDocument()
        return

    def removeHandler(self):
        self.writers[-1].endDocument()
        del self.writers[-1]
        return

    def pushResultTree(self, baseUri, implementation=None):
        """
        Start processing all content into a result-tree fragment.
        """
        writer = RtfWriter.RtfWriter(self.outputParams, baseUri,
                                     implementation or self._domimp)
        self.writers.append(writer)
        return

    def pushResultString(self):
        """
        Start processing creation events as string content.
        """
        writer = StringWriter.StringWriter(self.outputParams)
        self.writers.append(writer)
        return

    def pushResult(self, handler=None):
        '''
        Start processing all content into a separate result-tree handler.
        '''
        if handler is None:
            warnings.warn("Use pushResultTree(uri) to create RTFs",
                          DeprecationWarning, stacklevel=2)
            handler = RtfWriter.RtfWriter(self.outputParams,
                                          self.stylesheet.baseUri)
        self.writers.append(handler)
        handler.startDocument()
        return

    def popResult(self):
        '''End sub-result-tree and return any result'''
        handler = self.writers.pop()
        handler.endDocument()
        return handler.getResult()

    def output(self):
        """Returns the current output writer"""
        return self.writers[-1]

    def releaseRtf(self, rtf):
        # Even if Python is built without GC, our code assumes it is available
        warnings.warn("Not needed with Python's GC support",
                      DeprecationWarning, level=2)
        return

    def _stripElements(self, node):
        StripElements.StripElements(node, self.getStripElements())
        return

    def _normalizeParams(self, params):
        """
        params is a dictionary of top-level parameters.  The main task is to
        check this dictionary for lists of strings and convert these to
        a node set of text nodes
        """

        def to_unicode(s):
            try:
                # Try UTF-8
                return unicode(s, 'UTF-8')
            except ValueError:
                # Use encoding from locale
                try:
                    return unicode(s, DEFAULT_ENCODING)
                except ValueError:
                    raise Exception(
                        "String parameters must be Unicode objects or "
                        "strings encoded as UTF-8 or %s." %
                        DEFAULT_ENCODING)

        for k, v in params.items():
            if v and isinstance(v, types.StringType):
                params[k] = to_unicode(v)
            elif v and type(v) == type([]) and type(v[0]) in STRING_TYPES:
                doc = self._domimp.createRootNode(self.stylesheet.baseUri)
                nodeset = []
                for text in v:
                    if isinstance(text, types.StringType):
                        text = to_unicode(text)
                    nodeset.append(doc.createTextNode(text))
                params[k] = nodeset
        return

    def reclaim(self):
        return

    def reset(self):
        self.stylesheet = None
        #self._styReader = StylesheetReader.StylesheetReader(self._stylesheetIncPaths)
        return
